/*
 *      PSAPI library
 *
 *      Copyright 1998  Patrik Stridvall
 */

#include "winbase.h"
#include "windef.h"
#include "winerror.h"
#include "wine/debug.h"
#include "tlhelp32.h"
#include "psapi.h"

WINE_DEFAULT_DEBUG_CHANNEL(psapi);

#include <string.h>

/***********************************************************************
 *           EmptyWorkingSet (PSAPI.@)
 */
BOOL WINAPI EmptyWorkingSet(HANDLE hProcess)
{
  TRACE("(0x%08x)", hProcess);
  return SetProcessWorkingSetSize(hProcess, 0xFFFFFFFF, 0xFFFFFFFF);
}

/***********************************************************************
 *           EnumDeviceDrivers (PSAPI.@)
 */
BOOL WINAPI EnumDeviceDrivers(
  LPVOID *lpImageBase, DWORD cb, LPDWORD lpcbNeeded)
{
  FIXME("(%p, %ld, %p): stub\n", lpImageBase, cb, lpcbNeeded);

  if(lpcbNeeded)
    *lpcbNeeded = 0;

  return TRUE;
}


/***********************************************************************
 *           EnumProcesses (PSAPI.@)
 */
BOOL WINAPI EnumProcesses(DWORD *pProcessIds, DWORD cb, DWORD *pBytesReturned)
{
  HANDLE snapshot;
  DWORD count = 0, max = cb / sizeof(DWORD);
  BOOL ret;
  DWORD errcode;
  PROCESSENTRY32 pe32;

  TRACE("(%p, %ld, %p)\n", pProcessIds, cb, pBytesReturned);

  snapshot = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
  if( snapshot == INVALID_HANDLE_VALUE ) return FALSE;

  pe32.dwSize = sizeof( PROCESSENTRY32 );
  ret = Process32First( snapshot, &pe32 );

  TRACE("ret = %d, count = %ld\n", ret, count);
  while (ret && (count < max)) {
    pProcessIds[count] = pe32.th32ProcessID;
    TRACE("pid[%ld]=%ld, %s\n", count, pProcessIds[count], pe32.szExeFile);
    count++;
    
    ret = Process32Next( snapshot, &pe32 );
    TRACE("ret = %d, count = %ld\n", ret, count);
  }

  if( (errcode = GetLastError()) == ERROR_NO_MORE_FILES ) 
    SetLastError(NO_ERROR);
  TRACE("errcode=0x%08lx\n", errcode);
    
  CloseHandle( snapshot );

  if(pBytesReturned)
    *pBytesReturned = count * sizeof(DWORD);

  int i;
  for( i=0; i<count; i++ ) 
    TRACE("pProcessIds[%d]=0x%08lx\n", i, pProcessIds[i]);

  return TRUE;
}

/***********************************************************************
 *           EnumProcessModules (PSAPI.@)
 */
BOOL WINAPI EnumProcessModules(
  HANDLE hProcess, HMODULE *lphModule, DWORD cb, LPDWORD lpcbNeeded)
{
  BOOL ret;
  DWORD count = 0, actual = 0, max = cb / sizeof(HMODULE);
  DWORD errcode;
  HANDLE snapshot;
  MODULEENTRY32 me32;
  DWORD pid;

  TRACE("(hProcess=0x%08x, %p, %ld, %p)\n", hProcess, lphModule, cb, lpcbNeeded );

  pid = MapProcessHandle( hProcess );  /* FIXME: use GetProcessId if we implement it */
  TRACE("using pid=%ld\n", pid);
  snapshot = CreateToolhelp32Snapshot( TH32CS_SNAPMODULE, pid);
  if( snapshot == INVALID_HANDLE_VALUE ) return FALSE;

  me32.dwSize = sizeof( MODULEENTRY32 );
  ret = Module32First( snapshot, &me32 );

  TRACE("ret = %d, count = %ld\n", ret, count);
  while( ret ) {
    if( actual < max ) {
      lphModule[actual++] = me32.hModule;
    }
    TRACE("hModule[%ld]=0x%08lx, %s, %s, pid=%ld\n", count, me32.hModule, me32.szExePath, 
	                                         me32.szModule, me32.th32ProcessID);
    count++;
    
    ret = Module32Next( snapshot, &me32 );
    TRACE("ret = %d, count = %ld\n", ret, count);
  }

  if( (errcode = GetLastError()) == ERROR_NO_MORE_FILES ) 
    SetLastError(NO_ERROR);
  TRACE("errcode=0x%08lx\n", errcode);
    
  CloseHandle( snapshot );

  if(lpcbNeeded)
    *lpcbNeeded = count * sizeof(HMODULE);

  int i;
  for( i=0; i<actual; i++ ) 
    TRACE("hModule[%d]=0x%08x\n", i, lphModule[i]);
  
  return TRUE;
}

/***********************************************************************
 *          GetDeviceDriverBaseNameA (PSAPI.@)
 */
DWORD WINAPI GetDeviceDriverBaseNameA(
  LPVOID ImageBase, LPSTR lpBaseName, DWORD nSize)
{
  FIXME("(%p, %s, %ld): stub\n",
    ImageBase, debugstr_a(lpBaseName), nSize
  );

  if(lpBaseName && nSize)
    lpBaseName[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetDeviceDriverBaseNameW (PSAPI.@)
 */
DWORD WINAPI GetDeviceDriverBaseNameW(
  LPVOID ImageBase, LPWSTR lpBaseName, DWORD nSize)
{
  FIXME("(%p, %s, %ld): stub\n",
    ImageBase, debugstr_w(lpBaseName), nSize
  );

  if(lpBaseName && nSize)
    lpBaseName[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetDeviceDriverFileNameA (PSAPI.@)
 */
DWORD WINAPI GetDeviceDriverFileNameA(
  LPVOID ImageBase, LPSTR lpFilename, DWORD nSize)
{
  FIXME("(%p, %s, %ld): stub\n",
    ImageBase, debugstr_a(lpFilename), nSize
  );

  if(lpFilename && nSize)
    lpFilename[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetDeviceDriverFileNameW (PSAPI.@)
 */
DWORD WINAPI GetDeviceDriverFileNameW(
  LPVOID ImageBase, LPWSTR lpFilename, DWORD nSize)
{
  FIXME("(%p, %s, %ld): stub\n",
    ImageBase, debugstr_w(lpFilename), nSize
  );

  if(lpFilename && nSize)
    lpFilename[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetMappedFileNameA (PSAPI.@)
 */
DWORD WINAPI GetMappedFileNameA(
  HANDLE hProcess, LPVOID lpv, LPSTR lpFilename, DWORD nSize)
{
  FIXME("(hProcess=0x%08x, %p, %s, %ld): stub\n",
    hProcess, lpv, debugstr_a(lpFilename), nSize
  );

  if(lpFilename && nSize)
    lpFilename[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetMappedFileNameW (PSAPI.@)
 */
DWORD WINAPI GetMappedFileNameW(
  HANDLE hProcess, LPVOID lpv, LPWSTR lpFilename, DWORD nSize)
{
  FIXME("(hProcess=0x%08x, %p, %s, %ld): stub\n",
    hProcess, lpv, debugstr_w(lpFilename), nSize
  );

  if(lpFilename && nSize)
    lpFilename[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetModuleBaseNameA (PSAPI.@)
 */
DWORD WINAPI GetModuleBaseNameA(
  HANDLE hProcess, HMODULE hModule, LPSTR lpBaseName, DWORD nSize)
{
  FIXME("(hProcess=0x%08x, hModule=0x%08x, %s, %ld): stub\n",
    hProcess, hModule, debugstr_a(lpBaseName), nSize
  );

  if(lpBaseName && nSize)
    lpBaseName[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetModuleBaseNameW (PSAPI.@)
 */
DWORD WINAPI GetModuleBaseNameW(
  HANDLE hProcess, HMODULE hModule, LPWSTR lpBaseName, DWORD nSize)
{
  FIXME("(hProcess=0x%08x, hModule=0x%08x, %s, %ld): stub\n",
    hProcess, hModule, debugstr_w(lpBaseName), nSize);

  if(lpBaseName && nSize)
    lpBaseName[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetModuleFileNameExA (PSAPI.@)
 */
DWORD WINAPI GetModuleFileNameExA(
  HANDLE hProcess, HMODULE hModule, LPSTR lpFilename, DWORD nSize)
{
  HANDLE snapshot;
  MODULEENTRY32 me32;
  DWORD pid;
  BOOL ret;
  DWORD len = 0, errcode;

  TRACE("(hProcess=0x%08x,hModule=0x%08x, %p, %ld)\n",
        hProcess, hModule, lpFilename, nSize );

  if(lpFilename&&nSize)
    lpFilename[0]='\0'; 

  if( !hModule ) 
    FIXME("get EXE name of hProcess\n");

  pid = MapProcessHandle( hProcess );  /* FIXME: use GetProcessId if we implement it */
  TRACE("using pid=%ld\n", pid);
  snapshot = CreateToolhelp32Snapshot( TH32CS_SNAPMODULE, pid);
  if( snapshot == INVALID_HANDLE_VALUE ) return FALSE;

  me32.dwSize = sizeof( MODULEENTRY32 );
  ret = Module32First( snapshot, &me32 );

  TRACE("ret = %d\n", ret);
  while( ret ) {

    if( hModule == me32.hModule ) {
      TRACE("found hModule=0x%08lx, %s, %s, pid=%ld\n", 
	  me32.hModule, me32.szExePath, me32.szModule, me32.th32ProcessID);

      if(lpFilename && nSize) {
	strncpy(lpFilename, me32.szExePath, nSize);
	lpFilename[nSize-1] = '\0';
	len = strlen(lpFilename);
      }
      break;
    }
    
    ret = Module32Next( snapshot, &me32 );
    TRACE("ret = %d\n", ret);
  }

  if( (errcode = GetLastError()) == ERROR_NO_MORE_FILES ) 
    SetLastError(NO_ERROR);
  TRACE("errcode=0x%08lx\n", errcode);
    
  CloseHandle( snapshot );

  TRACE("returning filename='%s'\n", lpFilename);
  
  return len;
}

/***********************************************************************
 *           GetModuleFileNameExW (PSAPI.@)
 */
DWORD WINAPI GetModuleFileNameExW(
  HANDLE hProcess, HMODULE hModule, LPWSTR lpFilename, DWORD nSize)
{
  FIXME("(hProcess=0x%08x,hModule=0x%08x, %s, %ld): stub\n",
    hProcess, hModule, debugstr_w(lpFilename), nSize
  );

  if(lpFilename && nSize)
    lpFilename[0] = '\0';

  return 0;
}

/***********************************************************************
 *           GetModuleInformation (PSAPI.@)
 */
BOOL WINAPI GetModuleInformation(
  HANDLE hProcess, HMODULE hModule, LPMODULEINFO lpmodinfo, DWORD cb)
{
  HANDLE snapshot;
  MODULEENTRY32 me32;
  DWORD pid;
  BOOL ret, res;
  DWORD errcode;

  TRACE("(hProcess=0x%08x,hModule=0x%08x, %p, %ld)\n",
        hProcess, hModule, lpmodinfo, cb);

  memset(lpmodinfo, 0, cb);
  res = FALSE;

  if( !hModule ) 
    FIXME("get EXE name of hProcess\n");

  pid = MapProcessHandle( hProcess );  /* FIXME: use GetProcessId if we implement it */
  TRACE("using pid=%ld\n", pid);
  snapshot = CreateToolhelp32Snapshot( TH32CS_SNAPMODULE, pid);
  if( snapshot == INVALID_HANDLE_VALUE ) return FALSE;

  me32.dwSize = sizeof( MODULEENTRY32 );
  ret = Module32First( snapshot, &me32 );

  TRACE("ret = %d\n", ret);
  while( ret ) {

    if( hModule == me32.hModule ) {
      DWORD lfanew, entry;
      TRACE("found hModule=0x%08lx, %s, %s, pid=%ld\n", 
	  me32.hModule, me32.szExePath, me32.szModule, me32.th32ProcessID);

      lpmodinfo->lpBaseOfDll = me32.modBaseAddr;
      lpmodinfo->SizeOfImage = me32.modBaseSize;
      lpmodinfo->EntryPoint = NULL;
      if (!ReadProcessMemory(hProcess, &((IMAGE_DOS_HEADER*)me32.modBaseAddr)->e_lfanew,
                             &lfanew, sizeof(lfanew), NULL))
        return FALSE;
      if (!ReadProcessMemory(hProcess, &((IMAGE_NT_HEADERS*)(me32.modBaseAddr+lfanew))->OptionalHeader.AddressOfEntryPoint,
                             &entry, sizeof(entry), NULL))
        return FALSE;
      /* I'm unsure whether it's up to us or the caller to add the module base address...
       * Until proven otherwise, I'll believe it's the former. */
      lpmodinfo->EntryPoint = entry ? (me32.modBaseAddr + entry) : NULL;
      res = TRUE;
      break;
    }
    
    ret = Module32Next( snapshot, &me32 );
    TRACE("ret = %d\n", ret);
  }

  if( (errcode = GetLastError()) == ERROR_NO_MORE_FILES ) 
    SetLastError(NO_ERROR);
  TRACE("errcode=0x%08lx\n", errcode);
    
  CloseHandle( snapshot );

  TRACE("returning %d\n", res);
  
  return res;

}

/***********************************************************************
 *           GetProcessMemoryInfo (PSAPI.@)
 */
BOOL WINAPI GetProcessMemoryInfo(
  HANDLE Process, PPROCESS_MEMORY_COUNTERS ppsmemCounters, DWORD cb)
{
  FIXME("(hProcess=0x%08x, %p, %ld): stub\n",
    Process, ppsmemCounters, cb
  );

  memset(ppsmemCounters, 0, cb);

  return TRUE;
}

/***********************************************************************
 *           GetWsChanges (PSAPI.@)
 */
BOOL WINAPI GetWsChanges(
  HANDLE hProcess, PPSAPI_WS_WATCH_INFORMATION lpWatchInfo, DWORD cb)
{
  FIXME("(hProcess=0x%08x, %p, %ld): stub\n",
    hProcess, lpWatchInfo, cb
  );

  memset(lpWatchInfo, 0, cb);

  return TRUE;
}

/***********************************************************************
 *           InitializeProcessForWsWatch (PSAPI.@)
 */
BOOL WINAPI InitializeProcessForWsWatch(HANDLE hProcess)
{
  FIXME("(hProcess=0x%08x): stub\n", hProcess);

  return TRUE;
}

/***********************************************************************
 *           QueryWorkingSet (PSAPI.?)
 * FIXME
 *     I haven't been able to find the ordinal for this function,
 *     This means it can't be called from outside the DLL.
 */
BOOL WINAPI QueryWorkingSet(HANDLE hProcess, LPVOID pv, DWORD cb)
{
  FIXME("(hProcess=0x%08x, %p, %ld)\n", hProcess, pv, cb);

  if(pv && cb)
    ((DWORD *) pv)[0] = 0; /* Empty WorkingSet */

  return TRUE;
}
